DROP INDEX IF EXISTS idx_d_concept_id_low;
DROP INDEX IF EXISTS idx_dm_concept_id_low; -- see also: delete_normalized_concepts.sql
DROP INDEX IF EXISTS idx_dl_label_low;
DROP INDEX IF EXISTS idx_dx_xref_low;
DROP INDEX IF EXISTS idx_da_alias_low;
DROP INDEX IF EXISTS idx_d_as_association_low;
DROP INDEX IF EXISTS idx_rlv_concept_id_low;
